// finder.cpp
// Revision 12-mar-2004

#include "finder.h"

#include <vector>
#include <algorithm>

#include <ctype.h>
#include <regex.h>

#include "qtreassert.h"

//***********************************************
//	Auxiliary functions.
//***********************************************

namespace {

struct equal_nocase {
	bool operator () (const char c1, const char c2) const
	{
		return toupper (c1) == toupper (c2);
	}
};

} // namespace

//***********************************************
//		Finder
//***********************************************

Finder::string Finder::replace (const string & str, const string &)
{
	return str;
}

//***********************************************
//		StringFinder
//***********************************************

bool StringFinder::adjustfrom (const string & str, size_type & from)
{
	if (backwards () )
	{
		if (from <= searched.size () )
			return false;
		from-= searched.size () + 1;
	}
	else
	{
		++from;
		if (from >= str.size () )
			return false;
	}
	return true;
}

//***********************************************
//		NormalFinder
//***********************************************

Finder::size_type NormalFinder::find (const string & str)
{
	return str.find (searched);
}

Finder::size_type NormalFinder::find (const string & str, size_type from)
{
	if (! adjustfrom (str, from) )
		return npos;
	return str.find (searched, from);
}

//***********************************************
//		BackFinder
//***********************************************

Finder::size_type BackFinder::find (const string & str)
{
	return str.rfind (searched);
}

Finder::size_type BackFinder::find (const string & str, size_type from)
{
	if (! adjustfrom (str, from) )
		return npos;
	return str.rfind (searched, from);
}

//***********************************************
//		NocaseFinder
//***********************************************

Finder::size_type NocaseFinder::find (const string & str)
{
	string::const_iterator it=
		std::search (str.begin (), str.end (),
			searched.begin (), searched.end (),
			equal_nocase () );
	return (it == str.end () ) ?
		npos :
		std::distance (str.begin (), it);
}

Finder::size_type NocaseFinder::find (const string & str, size_type from)
{
	if (! adjustfrom (str, from) )
		return npos;
	string::const_iterator it=
		std::search (str.begin () + from, str.end (),
			searched.begin (), searched.end (),
			equal_nocase () );
	return (it == str.end () ) ?
		npos :
		std::distance (str.begin (), it);				
}

//***********************************************
//		BackNocaseFinder
//***********************************************

bool BackNocaseFinder::adjustfrom (const string & str, size_type & from)
{
	if (backwards () )
	{
		if (from <= searched.size () )
			return false;
		--from;
	}
	else
	{
		++from;
		if (from >= str.size () )
			return false;
	}
	return true;
}

Finder::size_type BackNocaseFinder::find (const string & str)
{
	string::const_iterator it=
		std::find_end (str.begin (), str.end (),
			searched.begin (), searched.end (),
			equal_nocase () );
	return (it == str.end () ) ?
		npos :
		std::distance (str.begin (), it);
}
Finder::size_type BackNocaseFinder::find (const string & str, size_type from)
{
	if (! adjustfrom (str, from) )
		return npos;
	string::const_iterator it=
		std::find_end (str.begin (), str.begin () + from,
			searched.begin (), searched.end (),
			equal_nocase () );
	return (it == str.begin () + from) ?
		npos :
		std::distance (str.begin (), it);
}

//***********************************************
//		WordBaseFinder
//***********************************************

WordBaseFinder::WordBaseFinder (const string & searched) :
	StringFinder (searched),
	ssize (searched.size () ),
	alnumfirst (! searched.empty () && isalnum (searched [0] ) ),
	alnumlast (! searched.empty () &&
		isalnum (searched [ssize - 1] ) )
{
}

bool WordBaseFinder::checkword (const string & str, size_type r)
{
	size_type end= r + ssize;
	return (! alnumfirst ||
		(r == 0 || ! isalnum (str [r - 1] ) ) ) &&
		(! alnumlast || 
		(end >= str.size () || ! isalnum (str [end] )
		) );
}

Finder::size_type WordBaseFinder::find (const string & str)
{
	return do_find (str, backwards () ? str.size () : 0);
}

Finder::size_type WordBaseFinder::find (const string & str, size_type from)
{
	if (! adjustfrom (str, from) )
		return npos;
	return do_find (str, from);
}

//***********************************************
//		WordFinder
//***********************************************

Finder::size_type WordFinder::do_find (const string & str, size_type from)
{
	size_type r;
	for (;;)
	{
		r= str.find (searched, from);
		if (r == npos)
			break;
		if (checkword (str, r) )
			break;
		from= r + 1;
	}
	return r;
}

//***********************************************
//		WordBackFinder
//***********************************************

Finder::size_type WordBackFinder::do_find (const string & str, size_type from)
{
	size_type r;
	for (;;)
	{
		r= str.rfind (searched, from);
		if (r == npos)
			break;
		if (checkword (str, r) )
			break;
		if (r == 0)
		{
			r= npos;
			break;
		}
		from= r - 1;
	}
	return r;
}

//***********************************************
//		WordNocaseFinder
//***********************************************

Finder::size_type WordNocaseFinder::do_find
	(const string & str, size_type from)
{
	size_type r;
	for (;;)
	{
		string::const_iterator it=
			std::search (str.begin () + from, str.end (),
				searched.begin (), searched.end (),
				equal_nocase () );
		r= (it == str.end () ) ?
			npos :
			std::distance (str.begin (), it);
		if (r == npos)
			break;
		if (checkword (str, r) )
			break;
		from= r + 1;
	}
	return r;
}

//***********************************************
//		WordBackNocaseFinder
//***********************************************

bool WordBackNocaseFinder::adjustfrom (const string & str, size_type & from)
{
	if (backwards () )
	{
		if (from <= searched.size () )
			return false;
		--from;
	}
	else
	{
		++from;
		if (from >= str.size () )
			return false;
	}
	return true;
}

Finder::size_type WordBackNocaseFinder::do_find
	(const string & str, size_type from)
{
	size_type r;
	for (;;)
	{
		string::const_iterator it=
			std::find_end (str.begin (),
				str.begin () + from,
				searched.begin (), searched.end (),
				equal_nocase () );
		r= (it == str.begin () + from) ?
			npos :
			std::distance (str.begin (), it);
		if (r == npos)
			break;
		if (checkword (str, r) )
			break;
		if (r == 0)
		{
			r= npos;
			break;
		}
		from= r - 1;
	}
	return r;
}


//***********************************************
//	RegexBaseFinder::Internal
//***********************************************

class RegexBaseFinder::Internal {
public:
	Internal (const string & searched, bool nocase);
	~Internal ();
	size_type find (const string & str);
	size_type find (const string & str, size_type from);
	size_type find_back (const string & str);
	size_type find_back (const string & str, size_type from);
	string replace (const string & str, const string & line);
	size_type length () const;
private:
	regex_t reg;
	std::vector <regmatch_t> pmatch;
	size_t pmlen;
	size_type desp;

	size_type do_find (const string & str, size_type from);
	void checkresult (int r);
};

RegexBaseFinder::Internal::Internal (const string & searched, bool nocase)
{
	int cflags= REG_EXTENDED;
	if (nocase)
		cflags|= REG_ICASE;
	int r= regcomp (& reg, searched.c_str (), cflags);
	checkresult (r);
	pmatch.resize (reg.re_nsub + 1);
	pmlen= pmatch.size ();
}

RegexBaseFinder::Internal::~Internal ()
{
	regfree (& reg);
}

Finder::size_type RegexBaseFinder::Internal::find
	(const string & str)
{
	return do_find (str, 0);
}

Finder::size_type RegexBaseFinder::Internal::find
	(const string & str, size_type from)
{
	++from;
	if (from >= str.size () )
		return npos;
	return do_find (str, from);
}

Finder::size_type RegexBaseFinder::Internal::do_find
	(const string & str, size_type from)
{
	int eflags= 0;
	if (from != 0)
		eflags|= REG_NOTBOL;
	int r= regexec (& reg, str.c_str () + from,
		pmlen, & pmatch [0], eflags);
	switch (r)
	{
	case 0:
		desp= 0;
		return pmatch [0].rm_so + from;
	case REG_NOMATCH:
		return Finder::npos;
	default:
		checkresult (r);
		return npos; // Make the compiler happy.
	}
}

Finder::size_type RegexBaseFinder::Internal::find_back
	(const string & str)
{
	int eflags= 0;
	const char * cstr= str.c_str ();
	int r= regexec (& reg, cstr, pmlen, & pmatch [0], eflags);
	if (r == REG_NOMATCH)
		return npos;
	checkresult (r);
	size_type res= pmatch [0].rm_so;
	eflags|= REG_NOTBOL;
	desp= 0;
	while ( (r= regexec (& reg, cstr + res + 1,
		pmlen, & pmatch [0], eflags) ) == 0)
	{
		desp= res + 1;
		res+= pmatch [0].rm_so + 1;
	}
	return res;
}

Finder::size_type RegexBaseFinder::Internal::find_back
	(const string & str, size_type from)
{
	int eflags= 0;
	const char * cstr= str.c_str ();
	int r= regexec (& reg, cstr, pmlen, & pmatch [0], eflags);
	if (r == REG_NOMATCH)
		return npos;
	checkresult (r);
	if (static_cast <size_t> (pmatch [0].rm_eo) >= from)
		return npos;
	size_type res= pmatch [0].rm_so;
	eflags|= REG_NOTBOL;
	desp= 0;
	while ( (r= regexec (& reg, cstr + res + 1,
		pmlen, & pmatch [0], eflags) ) == 0)
	{
		desp= res + 1;
		size_t endres= res + pmatch [0].rm_eo + 1;
		if (endres >= from)
			break;
		res+= pmatch [0].rm_so + 1;
	}
	return res;
}

Finder::string RegexBaseFinder::Internal::replace
	(const string & str, const string & line)
{
	size_type pos= 0;
	string nstr= str;
	while ( (pos= nstr.find ('$', pos) ) != npos)
	{
		if (pos >= str.size () - 1)
			break;
		char c= nstr [pos + 1];
		if (c == '$')
		{
			nstr.erase (pos, 1);
			++pos;
			continue;
		}
		if (! isdigit (c) )
		{
			pos+= 2;
			continue;
		}
		size_t n= c - '0';
		if (n > pmlen)
			nstr.erase (pos, 2);
		else
		{
			regoff_t ini= pmatch [n].rm_so;
			if (ini == regoff_t (-1) )
				nstr.erase (pos, 2);
			else
			{
				regoff_t l= pmatch [n].rm_eo - ini;
				nstr.replace (pos, 2, line, ini + desp, l);
				pos-= 2;
				pos+= l;
			}
		}
	}
	return nstr;
}

Finder::size_type RegexBaseFinder::Internal::length () const
{
	return pmatch [0].rm_eo - pmatch [0].rm_so;
}

void RegexBaseFinder::Internal::checkresult (int r)
{
	if (r != 0)
	{
		char buffer [256];
		regerror (r, & reg, buffer, sizeof (buffer) );
		throw std::string (buffer);
	}
}

//***********************************************
//		RegexBaseFinder
//***********************************************


RegexBaseFinder::RegexBaseFinder (const string & searched, bool nocase) :
	pin (new Internal (searched, nocase) )

{
}

RegexBaseFinder::~RegexBaseFinder ()
{
	delete pin;
}

Finder::size_type RegexBaseFinder::find (const string & str)
{
	return pin->find (str);
}

Finder::size_type RegexBaseFinder::find (const string & str, size_type from)
{
	return pin->find (str, from);
}

Finder::size_type RegexBaseFinder::length () const
{
	return pin->length ();
}

Finder::string RegexBaseFinder::replace
	(const string & str, const string & line)
{
	return pin->replace (str, line);
}

//***********************************************
//		RegexFinder
//***********************************************


RegexFinder::RegexFinder (const string & searched, bool nocase) :
	RegexBaseFinder (searched, nocase)
{
}

Finder::size_type RegexFinder::find (const string & str)
{
	return pin->find (str);
}

Finder::size_type RegexFinder::find (const string & str, size_type from)
{
	return pin->find (str, from);
}

//***********************************************
//		RegexBackFinder
//***********************************************


RegexBackFinder::RegexBackFinder (const string & searched, bool nocase) :
	RegexBaseFinder (searched, nocase)
{
}

Finder::size_type RegexBackFinder::find (const string & str)
{
	return pin->find_back (str);
}

Finder::size_type RegexBackFinder::find (const string & str, size_type from)
{
	return pin->find_back (str, from);
}

// End of finder.cpp
