#ifndef INCLUDE_TEXT_H
#define INCLUDE_TEXT_H

// text.h
// Revision 10-dec-2003

#include <string>
#include <vector>
#include <stdexcept>

class Col {
public:
	class Bad : public std::exception {
		const char * what () { return "Bad column"; }
	};

	Col () : n (0) { }
	explicit Col (size_t nn) : n (nn) { }
	size_t getn () const { return n; }
	Col & operator += (size_t inc)
	{
		if (n > colmax.n - inc)
			throw Bad ();
		n+= inc;
		return * this;
	}
	Col operator + (size_t nn) const
	{
		if (n > colmax.n - nn)
			throw Bad ();
		return Col (n + nn);
	}
	Col operator + (const Col & col) const
		{ return Col (n + col.n); }
	Col & operator ++ ()
	{
		if (n == colmax.n)
			throw Bad ();
		++n;
		return * this;
	}
	Col operator - (const Col & col) const
	{
		if (col.n > n)
			throw Bad ();
		return Col (n - col.n);
	}
	bool operator == (const Col & col) const
		{ return n == col.n; }
	bool operator != (const Col & col) const
		{ return n != col.n; }
	bool operator < (const Col & col) const
		{ return n < col.n; }
	bool operator <= (const Col & col) const
		{ return n <= col.n; }
	bool operator > (const Col & col) const
		{ return n > col.n; }
	static Col & max () { return colmax; }
	/*
	Col nexttab () const
		{ return Col ( (n / 8 + 1) * 8); }
	Col previoustab () const
		{ return Col ( (n / 8) * 8); }
	*/
	void nexttab ()
		{ n= (n / 8 + 1) * 8; }
	void previoustab ()
		{ n= (n / 8) * 8; }
	bool odd () const
		{ return n & 1; }
	friend std::ostream & operator << (std::ostream & os, const Col & col);
private:
	static Col colmax;
	size_t n;
};

inline std::ostream & operator << (std::ostream & os, const Col & col)
{
	os << col.n;
	return os;
}

class TextLine {
	std::string s;
public:
	TextLine () : s () {}
	TextLine (const std::string & snue) : s (snue) {}
	bool empty () const { return s.empty (); }
	void operator = (std::string snue)
		{ s= snue; }
	void operator += (char c)
		{ s+= c; }
	void operator += (const TextLine & tnue)
		{ s+= tnue.s; }
	inline char operator [] (size_t n) const { return s [n]; }
	size_t size () const { return s.size (); }
	const std::string & str () const { return s; }
	std::string substr (size_t n, size_t fin= std::string::npos) const
		{ return s.substr (n, fin); }
	size_t col2pos (Col col) const;
	Col pos2col (size_t pos) const;
	Col colmax () const;
	void deletechar (Col col);
	void deletecharpos (size_t pos);
	void insertchar (Col col, char c);
	void insertcharpos (size_t pos, char c);
	void insert (const std::string & str)
		{ s.insert (0, str); }
	void replace (size_t pos, std::string::size_type l,
			const std::string & str)
		{ s.replace (pos, l, str); }
	std::string wordat (Col col) const;
	friend std::ostream & operator << (std::ostream & os, const TextLine & tl);
};

class PosText {
public:
	PosText () : lin (0), pos (0) { }
	PosText (size_t nlin, size_t npos) : lin (nlin), pos (npos) { }
	size_t lin, pos;
	bool operator < (const PosText &pt) const
	{
		if (lin < pt.lin)
			return true;
		if (lin == pt.lin && pos < pt.pos)
			return true;
		return false;
	}
	bool operator == (const PosText &pt) const
		{ return lin == pt.lin && pos == pt.pos; }
	bool operator != (const PosText & pt) const
		{ return lin != pt.lin || pos != pt.pos; }
	void set (size_t nlin, size_t npos)
		{ lin= nlin; pos= npos; }
};

class Selection {
public:
	void init (int lin, int pos)
	{
		ptBegin.set (lin, pos);
		ptEnd.set (lin, pos);
	}
	const PosText & begin () { return ptBegin; }
	const PosText & end () { return ptEnd; }
	void movebegin (const PosText & npt) 
	{
		ptBegin= npt;
		adjust ();
	}
	void movebegin (int lin, int pos) { movebegin (PosText (lin, pos) ); }
	void moveend (const PosText & npt)
	{
		ptEnd= npt;
		adjust ();
	}
	void moveend (int lin, int pos) { moveend (PosText (lin, pos) ); }
private:
	PosText ptBegin, ptEnd;
	inline void adjust ()
		{ if (ptEnd < ptBegin) std::swap (ptBegin, ptEnd); }
};

class Text {
public:
	class Error : public std::runtime_error {
	public:
		Error (const std::string & str);
	};
	class ErrorSaving : public Error {
	public:
		ErrorSaving (const std::string & filename);
	};
	class ErrorReadOnly : public Error {
	public:
		ErrorReadOnly ();
	};

	typedef std::vector <TextLine> VecLine;

	Text ();
	#if 0
	static inline bool escapable (unsigned char c)
	{
		if (c < ' ' || c == '\x7F' || c == EscAlt) return true;
		if (c > (unsigned char) '\x7F' && c < (unsigned char) '\xA0')
			return true;
		return false;
	}
	#else
	static inline unsigned char escapable (unsigned char c)
	{
		if (c < ' ' || c == '\x7F') return '^';
		if (c > (unsigned char) '\x7F' && c < (unsigned char) '\xA0')
			return '\\';
		return '\0';
	}
	#endif
	std::string str () const;
	void c_str (char * str) const;
	bool copy (char * & str, size_t & len);
	static inline unsigned char escaped (unsigned char c)
	{
		if (c == '\x7F') return '?';
		//if (c == EscAlt) return '';
		if (c > (unsigned char) '\x7F' && c < (unsigned char) '\xA0')
			return c - '\x80' + '@';
		return c + '@';
	}
	/*
	static inline Col nexttab (Col col)
		//{ return (col / 8 + 1) * 8; }
		{ return col.nexttab (); }
	*/
	bool saved () { return isSaved; }
	bool hasname () { return ! filename.empty (); }
	std::string name () { return filename; }
	void clear ();
	bool load (const std::string & strFile);
	void save ();
	void save (const std::string & strFile);
	bool set (const char * str);
	size_t size () const;
	const TextLine & operator [] (size_t n) const
		{ return t [n]; }
	size_t lines () const { return nlines; }
	bool empty () const;
	void setline (size_t lin, const std::string & str);
	void deletecharpos (size_t lin, size_t pos);
	void insertchar (size_t lin, Col col, unsigned char c);
	void insertcharpos (size_t lin, size_t pos, unsigned char c);
	void append (size_t lin, const std::string & str);
	void appendchar (size_t lin, unsigned char c);
	size_t insertline (size_t lin, Col col, bool autoindent);
	void insertline (size_t lin, const std::string & str);
	void deleteline (size_t lin);
	void deletechar (size_t lin, size_t col);
	void joinline (size_t lin);
	void replace (size_t lin, size_t pos, std::string::size_type l,
			const std::string & str)
		{ t [lin].replace (pos, l, str); isSaved= false; }
	void push_back (const std::string & s)
		{ t.push_back (s); ++nlines; isSaved= false; }
	void modified (bool f) { isSaved= ! f; }
	bool modified () const { return ! isSaved; }
	void readonly (bool f) { isReadOnly= f; }
	bool readonly () const { return isReadOnly; }
	std::string wordat (size_t lin, Col col) const
	{ return t [lin].wordat (col); }
private:
	std::string filename;
	VecLine t;
	size_t nlines;
	static const unsigned char EscAlt= (unsigned char) '\x9B';
	bool isSaved;
	bool isReadOnly;

	void checkreadonly ();
};

#endif

// Fin de text.h
